<?php
/**
 * SearchModel class
 *
 * PHP version 5
 *
 * @category  Packages
 * @package   Ext.model
 * @author    Evgeniy Marilev <jeka@5-soft.com>
 * @copyright 2011 5-SOFT
 * @license   http://5-soft.com/packages/php/license proprietary
 * @link      http://5-soft.com/packages/php/docs
 */
Yii::import('packages.extensions.models.interfaces.ISearchModel');
/**
 * SearchModel is the base search model class used for searching data
 * 
 * @category  Packages
 * @package   Ext.model
 * @author    Evgeniy Marilev <jeka@5-soft.com>
 * @copyright 2011 5-SOFT
 * @license   http://5-soft.com/packages/php/license proprietary
 * @link      http://5-soft.com/packages/php/docs
 */
abstract class SearchModel extends CFormModel implements ISearchModel
{
    const SORT_MODE_ALPHABETICAL = 'alpha';
    const SORT_MODE_DATE = 'date';
    
    const DEFAULT_PAGE_SIZE = 20;
    
    const SORT_DIRECTION_ASC = 'asc';
    const SORT_DIRECTION_DESC = 'desc';
    
    /**
     * Count of records per page
     * @var integer
     */
    public $count;

    /**
     * Active results page index
     * @var integer
     */
    public $page;
    
    /**
     * Sort mode
     * @var string
     */
    public $sort;
    
    /**
     * Sort direction
     * @var string
     */
    public $direction;
    
    /**
     * Model class name used for searching
     * @var string
     */
    private $_modelClass;
    
    /**
     * Cached data provider
     * @var CDataProvider
     */
    private $_dataProvider;
    
    /**
     * Scopes cache
     * @var array
     */
    private $_scopes = array();
    
    /**
     * Returns validation rules map
     *
     * @return array validation rules
     * @see CModel::rules()
     */
    public function rules()
    {
        return array(
            array('sort', 'in', 'range' => $this->getSortModes()),
            array('direction', 'in', 'range' => $this->getSortDirections()),
            array('page, count', 'numerical', 'integerOnly' => true, 'min' => 0),
        );
    }
    
    /**
     * Initializes default values
     * 
     * @return void
     * @see CFormModel::init()
     */
    public function init()
    {
        parent::init();
        $this->count = $this->getDefaultPageSize();
        $this->direction = $this->getDefaultSortDirection();
        $this->sort = $this->getDefaultSortMode();
        $this->setScopes($this->getDefaultScopes());
    }
    
    /**
     * Returns default page size
     * 
     * @return integer default page size
     */
    public function getDefaultPageSize()
    {
        return self::DEFAULT_PAGE_SIZE;
    }
    
    /**
     * Returns default sort direction
     * 
     * @return mixed default sort direction
     */
    public function getDefaultSortDirection()
    {
        return false;
    }
    
    /**
     * Returns default sort mode
     * 
     * @return mixed default sort mode
     */
    public function getDefaultSortMode()
    {
        return false;
    }
    
    /**
     * Sets model class name used for searhing data
     * 
     * @param string $class model class name
     * 
     * @return void
     */
    public function setModelClass($class)
    {
        $this->_modelClass = $class;
    }
    
    /**
     * Returns model class name used for searching data
     * 
     * @return string model class name
     */
    public function getModelClass()
    {
        return $this->_modelClass;
    }
    
    /**
     * Returns available sort directions
     * 
     * @return array sort directions
     */
    public function getSortDirections()
    {
        return array(
            self::SORT_DIRECTION_ASC,
            self::SORT_DIRECTION_DESC,
            false,
        );
    }
    
    /**
     * Returns available sort modes
     * 
     * @return array sort modes
     */
    public function getSortModes()
    {
        return array(
            self::SORT_MODE_ALPHABETICAL,
            self::SORT_MODE_DATE,
            false,
        );
    }
    
    /**
     * Returns configured data provider
     * 
     * @return CDataProvider configured data provider
     * @throws Exception
     */
    public function search()
    {
        if (!isset($this->_dataProvider)) {
            if ($this->validate()) {
                $dataProvider = $this->build();
                $dataProvider->pagination = $this->getPagination();
                if ($dataProvider instanceof CActiveDataProvider) {
                    $criteria = $dataProvider->criteria;
                    if (empty($criteria->order)) {
                        $this->applySort($criteria);
                    }
                }
                $this->_dataProvider = $dataProvider;
            } else {
                $errors = $this->getErrors();
                $this->displayErrors($errors);
            }
        }
        return $this->_dataProvider;
    }
    
    /**
     * Sets active scopes set
     * 
     * @param mixed $scopes scopes set (comma separated string or array)
     * 
     * @return void
     */
    public function setScopes($scopes)
    {
        if (is_string($scopes)) {
            $scopes = explode(',', $scopes);
        } else if (!is_array($scopes)) {
            throw new Exception('Invalid scopes expression.');
        }
        $this->_scopes = $scopes;
    }
    
    /**
     * Adds new scope
     * 
     * @param string $scope scope
     * 
     * @return void
     */
    public function addScope($scope)
    {
        if (!in_array($scope, $this->_scopes)) {
            $this->_scopes[] = $scope;
        }
    }
    
    /**
     * Returns active scopes set
     * 
     * @return array scopes set
     */
    public function getScopes()
    {
        return $this->_scopes;
    }
    
    /**
     * Returns default scopes
     * 
     * @return array default scopes
     */
    public function getDefaultScopes()
    {
        return array();
    }
    
    /**
     * Returns whether model has active scope with
     * specified name
     * 
     * @param string $name scope name
     * 
     * @return boolean whether scope is active
     */
    public function hasScope($name)
    {
        return in_array($name, $this->_scopes);
    }
    
    /**
     * Applies sorting conditions for criteria
     * 
     * @param CDbCriteria $criteria search criteria
     * 
     * @return void
     */
    protected function applySort($criteria)
    {
        if (!empty($criteria->order) && $this->direction !== false) {
            $criteria->order .= ' ' . $this->direction;
        }
    }
    
    /**
     * Returns pagination configuration
     * 
     * @return array pagination config
     */
    protected function getPagination()
    {
        if (!$this->count) {
            return false;
        } else {
            $ret = array(
                'pageSize' => $this->count,
            );
            if (isset($this->page)) {
                $ret['currentPage'] = $this->page <= 0 ? 0 : $this->page - 1;
            }
            return $ret;
        }
    }
    
    /**
     * Displays validation errors
     * 
     * @param array &$data errors map
     * 
     * @return void
     * @throws Exception
     */
    protected function displayErrors(&$data)
    {
        $errors = array();
        foreach ($data as $name => $error) {
            $errors = array_merge($errors, $error);
        }
        $params = array('{ERRORS}' => implode($errors));
        $message = Yii::t('assert', 'There are following model validation errors: "{ERRORS}"', $params);
        throw new Exception($message);
    }
    
    /**
     * Builds data provider used for fetching data
     * 
     * @return CDataProvider configured data provider
     */
    protected abstract function build();
    
    /**
     * Asserts whether model object is instance of search model
     * 
     * @param object $model model instance
     * 
     * @return void
     * @throws Exception
     */
    public static function assertSearchModel($model)
    {
        if (!$model instanceof ISearchModel) {
            $params = array(
                '{CLASS}' => get_class($model),
                '{INTERFACE}' => 'ISearchModel',
            );
            $message = Yii::t('assert', 'The class "{CLASS}" must implements "{INTERFACE}"!', $params);
            throw new Exception($message);
        }
    }
}