<?php
/**
 * Sandbox main model behavior.
 *
 * Should be attached to the revision model:
 *    public function behaviors()
 *    {
 *         return array(
 *             'SandboxBehavior' => array(
 *                 'class' => 'application.models.sandbox.SandboxBehavior',
 *                 'hasRevisionHistory' => true,
 *                 'revisionModel' => 'PublisherProfileRevision',
 *             ),
 *         );
 *     }
 * (@link revisionModel} parameter is required - the name of the revision model class
 * {@link hasRevisionHistory} parameter allows to enable (true) or disable (false) intermediate revisions history
 *
 * @author seb
 * @package app.models
 * @subpackage app.models.sandbox
 */
class SandboxBehavior extends CActiveRecordBehavior {

    const SANDBOX_REVISION = 0;
    const HISTORY_REVISION = 1;
    const PUBLISHED_REVISION = 2;
    
    const PUBLISH_ERROR = 0;
    const PUBLISH_SUCCESS = 1;
    const PUBLISH_NO_DATA_CHANGED = 2;

    /**
     *
     * Whether to support revisions history
     * false - only sandbox & published revisions
     * true - sandbox, published and history revisions
     * @var boolean
     */
    public $hasRevisionHistory = false;

    /**
     * Revision model name
     * @var CActiveRecord
     */
    public $revisionModel;

    /**
     * Relation revision model foreign key
     * @var string
     */
    public $revisionKey;

    /**
     * Field name where revision type is stored
     * @var string
     */
    public $revisionTypeField = 'revision_type';
    
    /**
     * Whether require create published revision when sandbox data not changed
     * @var bool
     */
    public $publishIfNoDataChanged = false;

    /**
     * Attach required relations
     */
    public function attach($owner) {
        parent::attach($owner);
        $owner->getMetaData()->addRelation('sandbox', array(
            CActiveRecord::HAS_ONE, $this->revisionModel, $this->revisionKey,
                'condition'=>'sandbox.' . $this->revisionTypeField . ' = ' . self::SANDBOX_REVISION
        ));
        $owner->getMetaData()->addRelation('published', array(
            CActiveRecord::HAS_ONE, $this->revisionModel, $this->revisionKey,
                'condition'=>'published.' . $this->revisionTypeField . ' = ' . self::PUBLISHED_REVISION
        ));
        if ($this->hasRevisionHistory) {
            $owner->getMetaData()->addRelation('history_revisions', array(
                CActiveRecord::HAS_MANY, $this->revisionModel, $this->revisionKey,
                    'condition'=>'history_revisions.' . $this->revisionTypeField . ' = ' . self::HISTORY_REVISION
            ));
        }
    }

    /**
     * Find and return history revisions and published revision
     * @return array revision objects
     */
    /*public function findHistory()
    {
        $revision = $this->revisionModel();

        $criteria = new CDbCriteria;
        $criteria->addCondition($revision->{$this->revisionKey} . '=' . $this->owner->id);
        $criteria->addInCondition($revision->revisionTypeField,
            array(SandboxBehavior::HISTORY_REVISION,SandboxBehavior::PUBLISHED_REVISION)
        );

        return $revision->findAll($criteria);
    }*/

    /**
     * Try to publish sandbox revision to published revision and return publishing result
     * @return int publishing operation result 
     */
    public function publish()
    {
        if (!$this->publishIfNoDataChanged) {
            $changed = $this->sandboxChanged();
            if (!$changed) {
                return self::PUBLISH_NO_DATA_CHANGED;
            }
        }
        if ($this->beforePublish()) {
            $publishResult = $this->hasRevisionHistory 
                       ? $this->publishWithHistory() 
                       : $this->publishSandbox();
            $this->afterPublish();
            return $publishResult;
        }
        return self::PUBLISH_ERROR;
    }
    
    /**
     * Checks sandbox for any changes
     * @return bool whether sandbox was changed
     */
    public function sandboxChanged() {
        $ret = false;
        $published = $this->owner->published;
        if ($published) {
            $sandbox = $this->owner->sandbox;
            return !$sandbox->compareWith($published, array($this->revisionTypeField)); 
        } else {
            return true;
        }
        return $ret;
    } 

    /**
     * Cancel sandbox and replace it with published revision content
     */
    public function cancel()
    {
        $published = $this->owner->published;
        if (!$published) {
            return false;
        } else {
            $this->copyRevision($published, $this->owner->sandbox, array($this->revisionTypeField));
            $this->afterCancel();
            return true;
        }
    }

    /**
     * Revert sandbox to the specified revision
     */
    public function revert($revisionId)
    {
        $revision = $this->findRevision($revisionId);
        if (!$revision) {
            return false;
        } else {
            $this->copyRevision($revision, $this->owner->sandbox, array($this->revisionTypeField));
            return true;
        }
    }

    /**
     * This event is raised before publishing sandbox.
     * @param CEvent the event parameter
     * @see beforePublish
     */
    public function onBeforePublish($event)
    {
        $this->raiseEvent('onBeforePublish',$event);
    }

    /**
     * This event is raised after the sandbox is published.
     * @param CEvent the event parameter
     */
    public function onAfterPublish($event)
    {
        $this->raiseEvent('onAfterPublish',$event);
    }

    /**
     * This event is raised after the sandbox is canceled.
     * @param CEvent the event parameter
     */
    public function onAfterCancel($event)
    {
        $this->raiseEvent('onAfterCancel',$event);
    }

    /**
     * Find specified revision
     * @param integer $revisionId revision id to find
     * @return revision object or null if revision not found
     */
    public function findRevision($revisionId)
    {
        $model = $this->revisionModel();
        return $model->findByPk($revisionId);
    }
    
    /**
     * This method is invoked before publishing a sandbox.
     * The default implementation raises the {@link onBeforePublish} event.
     * You may override this method to do any work before publishing.
     * Make sure you call the parent implementation so that the event is raised properly.
     * @return boolean whether the publishing should be executed. Defaults to true.
     */
    protected function beforePublish()
    {
        if($this->hasEventHandler('onBeforePublish'))
        {
            $event=new CModelEvent($this);
            $this->onBeforePublish($event);
            return $event->isValid;
        }
        else
            return true;
    }

    /**
     * This method is invoked after publishing a sandbox successfully.
     * The default implementation raises the {@link onAfterPublish} event.
     * You may override this method to do postprocessing after publishing.
     * Make sure you call the parent implementation so that the event is raised properly.
     */
    protected function afterPublish()
    {
        if($this->hasEventHandler('onAfterPublish'))
            $this->onAfterPublish(new CEvent($this));
    }

    /**
     * This method is invoked after sandbox changes are canceled.
     * The default implementation raises the {@link onAfterCancel} event.
     * You may override this method to do postprocessing after canceling.
     * Make sure you call the parent implementation so that the event is raised properly.
     */
    protected function afterCancel()
    {
        if($this->hasEventHandler('onAfterCancel'))
            $this->onAfterCancel(new CEvent($this));
    }

     /**
      * Publish sandbox with history.
      * Old published revision becomes history, sandbox is copied to new published revision
      */
     private function publishWithHistory()
     {
         $newRevision = $this->createRevision(SandboxBehavior::PUBLISHED_REVISION);
         $oldPublished = $this->owner->published;
         //if (!$oldPublished) $oldPublished = $this->findPublished();

         if ($oldPublished) {
            $oldPublished->{$this->revisionTypeField} = SandboxBehavior::HISTORY_REVISION;
            if (!$oldPublished->save(false)) {
                throw new CException("Can not save history revision. " . CHtml::errorSummary($oldPublished));
            }
         }

         $ret = $this->copyRevision($this->owner->sandbox, $newRevision, array($this->revisionTypeField));
         $ret = $ret ? self::PUBLISH_SUCCESS : self::PUBLISH_ERROR;
          return $ret;
     }

    /**
     * Publish sandbox without history
     * Sandbox is copied to the published revision
     */
    private function publishSandbox()
    {
        //we should do this because AppSandboxBehavior subclass
        //returns null is publisher is not active
        //not very good solution, but to fix we need to refactor many other code
        //(do not return null in AppSandboxBehavior, but check publisher
        // status anywhere we want to get published data)
        $published = $this->owner->published;
        if (!$published) {
            $published = $this->createRevision(SandboxBehavior::PUBLISHED_REVISION);
        }
        $ret = $this->copyRevision($this->owner->sandbox, $published, array($this->revisionTypeField));
        $ret = $ret ? self::PUBLISH_SUCCESS : self::PUBLISH_ERROR;
        return $ret;
    }

    /**
     * Copy one revision to another.
     * Revision class should have CopyBehavior
     * @param CActiveRecord $from revision to copy from
     * @param CActiveRecord $to revision to copy to
     */
    public function copyRevision($from, $to, $skipAttributes = array())
    {
        return $to->copyFrom($from, $skipAttributes);
        if ($to->asa('CopyBehavior')) {
            return $to->copyFrom($from);
        } else {
            $revClass = get_class($to);
            throw new CException("Revision class $revClass does not have CopyBehavior");
        }
    }

    /**
     * After construct event handler - creates sandbox revision
     * @param CEvent $event
     */
    public function afterConstruct($event)
    {
        $this->owner->sandbox = $this->createRevision(SandboxBehavior::SANDBOX_REVISION);
    }

    /**
     * After save event handler
     * Saves sandbox revision
     */
    public function afterSave($event) {
        if ($this->owner->sandbox->getIsNewRecord()) {
            $this->owner->sandbox->{$this->revisionKey} = $this->owner->id;
            if (!$this->owner->sandbox->save(false)) {
                throw new CException("Can not save sandbox. " . CHtml::errorSummary($this->owner->sandbox));
            }
        }
    }

    /**
     * Calls model() method from revision class
     */
    private function revisionModel() {
        return call_user_func(array($this->revisionModel, 'model'));
    }

    /**
     * Create new revision
     * @param integer $revisionType new revision type
     */
    private function createRevision($revisionType = null) {
        $revision = new $this->revisionModel;
        if (isset($revisionType)) {
            $revision->{$this->revisionKey} = $this->owner->id;
            $revision->{$this->revisionTypeField} = $revisionType;
        }
        return $revision;
    }
}
