<?php
/**
 * CActiveTreeDataProvider class
 *
 * PHP version 5
 * 
 * @category   Widgets
 * @package    Ext.widget
 * @subpackage Ext.widget.treeview
 * @author     Evgeniy Marilev <jeka@5-soft.com>
 */
/**
 * CActiveTreeDataProvider is the dataprovider class
 * for fetching tree-structure based data (adjacency model)
 * 
 * @category   Widgets
 * @package    Ext.widget
 * @subpackage Ext.widget.treeview
 * @author     Evgeniy Marilev <jeka@5-soft.com>
 */
class CActiveTreeDataProvider extends CActiveDataProvider
{
    const KEY_ID = 'id';
    const KEY_TEXT = 'text';
    const KEY_HAS_CHILDREN = 'hasChildren';
    const KEY_CHILDREN = 'children';
    
    /**
     * Root node id
     * @var integer
     */
    public $rootNodeId;
    
    /**
     * Tree data cache
     * @var array
     */
    protected $_treeData;
    
    /**
     * Tree data making options
     * @var array
     */
    public $treeDataOptions;
    
    /**
     * Constructs class based object
     * 
     * @param string $modelClass model's class name
     * @param array  $config     provider's config
     * 
     * @return void
     */
    public function __construct($modelClass, $config=array())
    {
        parent::__construct($modelClass, $config);
        $this->treeDataOptions = isset($this->treeDataOptions) ? $this->treeDataOptions : array();
        $this->treeDataOptions = CMap::mergeArray(array(
            'keys' => array(
                self::KEY_ID => 'id',
                self::KEY_TEXT => 'text',
                self::KEY_HAS_CHILDREN => 'hasChildren',
                self::KEY_CHILDREN => 'children',
            ),
        ), $this->treeDataOptions);
    }
    
    /**
     * Fetchs raw data
     * 
     * @return array fetched data
     * @see CActiveDataProvider::fetchData()
     */
    protected function fetchData()
    {
        $criteria = clone $this->getCriteria();

        if (($pagination = $this->getPagination()) !== false) {
            $pagination->setItemCount($this->getTotalItemCount());
            $pagination->applyLimit($criteria);
        }

        if (($sort = $this->getSort()) !== false) {
            $sort->applyOrder($criteria);
        }

        $baseCriteria = $this->model->getDbCriteria(false);
        if ($baseCriteria !== null) {
            $baseCriteria = clone $baseCriteria;
        }
        $tableName = $this->model->tableName();
        $data = Yii::app()->db->getCommandBuilder()->createFindCommand($tableName, $criteria)->queryAll();
        $this->model->setDbCriteria($baseCriteria);
        return $data;
    }
    
    /**
     * Returns parent tree item
     * 
     * @param array tree item
     * 
     * @return array root node item
     */
    public function getParentNodeItem($treeItem)
    {
        if (isset($treeItem['parent_id'])) {
            $criteria = new CDbCriteria;
            $criteria->compare('id', $treeItem['parent_id']);
            $tableName = $this->model->tableName();
            return Yii::app()->db->getCommandBuilder()->createFindCommand($tableName, $criteria)->queryRow();
        }
        return null;
    }
    
    /**
     * Returns data in treeview control's format
     * 
     * @return array tree data
     */
    public function getTreeData()
    {
        if (!isset($this->_treeData)) {
            $levels = array();
            $items = $this->getData();
            
            $minLevel = null;
            $minLevelParent = null;
            foreach ($items as $item) {
                $dataItem = $this->getTreeItemData($item);
                if ($dataItem === null) {
                    continue;
                }
                $parentKey = $item['parent_id'] === null ? 'null' : $item['parent_id'];
                if (!isset($levels[$item['level']])) {
                    $levels[$item['level']] = array();
                }
                $levels[$item['level']][$parentKey][] = $dataItem;
                if (!isset($minLevel) || $item['level'] < $minLevel) {
                    $minLevel = $item['level'];
                    $minLevelParent = $item['parent_id'];
                }
            }
            
            $levelKeys = array_keys($levels);
            if (empty($levelKeys)) {
                $this->_treeData = array();
            } else {
                $parent = isset($minLevelParent) ? $minLevelParent : 'null';
                $this->_treeData = $this->getTreeDataRec($levels, $parent, $minLevel);
            }
            
            $this->postFilterTreeData();
        }
        
        return $this->_treeData;
    }
    
    /**
     * Postfilters tree data
     * 
     * @return void
     */
    protected function postFilterTreeData()
    {
        $curItem = &$this->_treeData;
        $path = array();
        $current = $this->rootNodeId ? $this->getParentNodeItem(array('parent_id' => $this->rootNodeId)) : null;
        while ($current) {
            $path[] = $current['id'];
            $current = $this->getParentNodeItem($current);
        }
        $path = array_reverse($path);
        
        $count = count($path);
        $childrenKey = $this->treeDataOptions['keys'][self::KEY_CHILDREN];
        for ($i = 0; $i < $count; ++$i) {
            foreach ($curItem as $j => $item) {
                if ($item['id'] == $path[$i]) {
                    if (isset($curItem[$j][$childrenKey])) {
                        $curItem = &$curItem[$j][$childrenKey];
                    } else {
                        $curItem = array();
                    }
                    break;
                }
            }
        }
        
        $this->_treeData = $curItem;
    }
    
    /**
     * Returns tree item data by record
     * 
     * @param array $item tree item
     * 
     * @return array tree item data
     */
    public function getTreeItemData($item)
    {
        $keys = $this->treeDataOptions['keys'];
        return array(
            $keys[self::KEY_ID] => $item['id'],
            $keys[self::KEY_TEXT] => $item['name'],
            $keys[self::KEY_HAS_CHILDREN] => $item['is_leaf'] == 0,
        );
    }
    
    /**
     * Returns data in tree's data recursively
     * 
     * @param array   $data   data array
     * @param string  $parent parent element id
     * @param integer $level  item level
     * 
     * @param array $data
     */
    protected function getTreeDataRec(&$data, $parent, $level)
    {
        if (!isset($data[$level]) || !isset($data[$level][$parent])) {
            return array();
        }
        $items = &$data[$level][$parent];
        $ret = array();
        foreach ($items as $i => $item) {
            $ret[$i] = $item;
            if ($item[$this->treeDataOptions['keys'][self::KEY_HAS_CHILDREN]]) {
                $ret[$i][$this->treeDataOptions['keys'][self::KEY_CHILDREN]] = $this->getTreeDataRec($data, $item['id'], $level + 1);
            }
        }
        return $ret;
    }
}